/*
  file:   mod_auth_any.c
  author: Nafees Bin Zafar <binzafar@musc.edu>
  date:   7/12/00

  $Id: mod_auth_any.c,v 1.41 2004/11/22 20:06:52 satyap Exp $
  
  See docs/CHANGELOG in the source distribution.
  
*/

/* ====================================================================
 * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * http_auth: authentication
 * 
 * Rob McCool
 * 
 * Adapted to Apache by rst.
 *
 * dirkx - Added Authoritative control to allow passing on to lower
 *         modules if and only if the user-id is not known to this
 *         module. A known user with a faulty or absent password still
 *         causes an AuthRequired. The default is 'Authoritative', i.e.
 *         no control is passed along.
 */

#include "mod_auth_any.h"


#ifdef WITH_APACHE2
typedef struct {
#else
typedef struct auth_config_struct {
#endif
  char *auth_pwfile;
  char *auth_clparams;
  char *auth_grpfile;
  MAA_POOL *pool;
  int auth_authoritative;
  int auth_cache;
} auth_config_rec;

static apr_thread_mutex_t *mx;

void child_exec(request_rec *r,
        char* user,
        const char* password,
        char *auth_pwfile,
        char *auth_clparams,
        int fd[]
        ) {
    int i;
    char *buf;
    char **cl_args;
        ap_log_rerror (MAA_DEBUG,
                "mod_auth_any.child_exec, entry");

#ifdef __sun
    snprintf (m_envstr, sizeof(m_envstr), "%s=%s", "REMOTE_ADDR",
            r -> connection -> remote_ip);
    putenv (m_envstr);
#else
    setenv ("REMOTE_ADDR", r -> connection -> remote_ip, 1);
#endif
    /* close open fds */
        ap_log_rerror (MAA_DEBUG,
                "mod_auth_any.child_exec, closing open fds");
    i = sysconf (_SC_OPEN_MAX);
    while (i >= 0) {
        if (i != fd[1]) { /* FIXME i!=1  ? */
            fcntl (i, F_SETFD, FD_CLOEXEC);
        }
        i--;
    }
        ap_log_rerror (MAA_DEBUG,
                "mod_auth_any.child_exec, make the write end our stdout");
    /* make the write end of the pipe our stdout */
    if (fd[1] != STDOUT_FILENO) {
        close (STDOUT_FILENO);
        dup2 (fd[1], STDOUT_FILENO);
        close (STDERR_FILENO);
        dup2 (fd[1],STDERR_FILENO);
        fcntl (STDOUT_FILENO, F_SETFD, 0);
        fcntl (fd[1], F_SETFD, FD_CLOEXEC);
    }
        ap_log_rerror (MAA_DEBUG,
                "mod_auth_any.child_exec, alloc buffers");
    if ((buf=(char *)malloc((strnlen(auth_pwfile,   MAA_BUFLEN)+
                        strnlen(auth_clparams, MAA_BUFLEN)+
                        strnlen(user,          MAA_BUFLEN)+
                        strnlen(password,      MAA_BUFLEN)+
                        4) * 
                    /* 4 for terminating nulls -- but
                       that's not necessary, i wonder why 
                       we had it? oh well, we only waste 4 
                       bytes if we leave it in. --satya */
                    sizeof(char))) == NULL) {
        ap_log_rerror (MAA_ERROR,
                "mod_auth_any.child_exec, unable to allocate space for temporary buffer");
    /*    return NULL;*/
        _exit(1);
    }
        ap_log_rerror (MAA_DEBUG,
                "mod_auth_any.child_exec, manipulate buffers");

    strncpy(buf,auth_pwfile,strnlen(auth_pwfile,MAA_BUFLEN)+1);
    if (strnlen(auth_clparams,257) > 0) {
        strncat(buf," ",2);
        strncat(buf,auth_clparams,strnlen(auth_clparams,MAA_BUFLEN));
    }
    strncat(buf," ",2);
    strncat(buf,user,strnlen(user,MAA_BUFLEN));
    strncat(buf," ",2);
    strncat(buf,password,strnlen(password,MAA_BUFLEN));
        ap_log_rerror (MAA_DEBUG,
                "mod_auth_any.child_exec, build clargs");

    cl_args = split(' ',buf);
    free(buf);
        ap_log_rerror (MAA_DEBUG,
                "mod_auth_any.child_exec, about to execv");

    execv (auth_pwfile, cl_args);
        ap_log_rerror (MAA_ERROR,
                "mod_auth_any.child_exec, can't be here");
    _exit (1); /* FIXME why is there code after this? */
        ap_log_rerror (MAA_ERROR,
                "mod_auth_any.child_exec, after _exit. something's badly wrong.");
    while(*cl_args != NULL) {
        free(*cl_args);
        cl_args++;
    }
} /* child_exec */


static void *create_auth_dir_config(MAA_POOL *p, char *d)
{
  int threaded;
  auth_config_rec *sec =
    (auth_config_rec *) MAA_PCALLOC (p, sizeof(auth_config_rec));
  sec->auth_pwfile = NULL;	/* just to illustrate the default really */
  sec->auth_clparams = ""; /* no params by default */
  sec->auth_grpfile = NULL;	/* unless you have a broken HP cc */
  sec->auth_authoritative = 1;	/* keep the fortress secure by default */
  sec->auth_cache = 0; /* no caching by default */
  sec->pool=p;

  mx=NULL;

#ifdef WITH_APACHE2
  ap_mpm_query(AP_MPMQ_IS_THREADED,&threaded);
  if(threaded) {
      apr_thread_mutex_create(&mx,APR_THREAD_MUTEX_DEFAULT,p);
  }
#else
  mx=ap_create_mutex(NULL);
#endif
  threaded=0;
  
  return sec;
}

static const char *set_auth_slot(cmd_parms *cmd, void *offset, const char *f, const char *t)
{
  if (t && strcmp(t, "standard"))
    return MAA_STRCAT (cmd->pool, "Invalid auth file type: ", t, NULL);

  return ap_set_file_slot(cmd, offset, (char *)f);
}


static const command_rec auth_any_cmds[] =
  {
      MAA_AP_INIT("AuthAnyUserProg",auth_pwfile,"The external authenticator program"),
      MAA_AP_INIT("AuthAnyGroupFile",auth_grpfile,"text file containing group names and member user IDs"),
      MAA_AP_INIT("AuthAnyCLParams",auth_clparams,"First parameters send to the script"),
      MAA_AP_FLAG("AuthAnyAuthoritative",auth_authoritative,"Set to 'off' to allow access control to be passed along to lower modules if the UserID is not known to this module"),
      MAA_AP_FLAG("AuthAnyUseCaching",auth_cache,"Set to 'off' to disable caching, 'on' to enable. Default: off"),
     {NULL}
  };

#ifdef WITH_APACHE2
module AP_MODULE_DECLARE_DATA auth_any_module;
#else
module MODULE_VAR_EXPORT auth_any_module;
#endif




/*
 *  Call the external program auth_pwfile with user and password as its
 *  arguments.  These should have been passed in over a pipe, but it's too
 *  late to change that now.  The called program should print the name of the
 *  valid user to stdout, or print "Authentication Error" on failure.
 */
static char *get_pw(
        request_rec* r,
        char* user,
        const char* password,
        auth_config_rec *sec) {
  char *auth_pwfile;
  char *auth_clparams;
  int auth_cache;
  char result[MAA_BUFLEN];
  char tmpbuf[4];
  char *buf;
  int fd[2], i, length = 0;
  char *found;
  char *encpass=NULL;
  pid_t pid;
  extern int errno;
#ifdef __sun
  static char m_envstr[MAA_BUFLEN]; /* sun's putenv() call is a bunch of bull malarkey */
#endif

  memset(result, '\0', MAA_BUFLEN);
  auth_pwfile=sec->auth_pwfile;
  auth_clparams=sec->auth_clparams;
  auth_cache=sec->auth_cache;
  if(MAA_DEBUGLVL>0) 
      ap_log_rerror(MAA_DEBUG,"mod_auth_any authenticating %s",user);

  if(auth_cache!=0) { /* if caching not disabled */
#ifdef MAA_USE_MD5
#ifdef WITH_APACHE2
      encpass=(char *)apr_palloc(sec->pool,APR_MD5_DIGESTSIZE);
      if(encpass==NULL) {
          ap_log_rerror(MAA_DEBUG,"Out of memory allocating enough for encrypted password");
          return NULL;
      }
      apr_md5(encpass,password,(apr_size_t)strnlen(password,MAA_BUFLEN));
#else
      encpass=ap_md5(sec->pool,password);
#endif
#endif
      if(MAA_DEBUGLVL>0)
          ap_log_rerror(MAA_DEBUG,"looking %s up in cache",user);
      /* lookup user in cache */
      found=NULL;
      TMPM_LOCK();
      found=lookup_user(r,user,MAA_PASS);
      if(MAA_DEBUGLVL>4)
          ap_log_rerror(MAA_DEBUG,"after lookup_user, found=%s up in cache",found);
      TMPM_UNLOCK();
      if(found!=NULL) {
          if(MAA_DEBUGLVL>0)
              ap_log_rerror(MAA_DEBUG,"%s found in cache",user);
          return strdup(found);
      }
      if(MAA_DEBUGLVL>0)
          ap_log_rerror(MAA_DEBUG,"%s back from lookup in cache",user);
  }

    
  /* sanity check the auth_pwfile */
  if ((auth_pwfile == NULL) || (auth_pwfile[0] != '/')) {
    ap_log_rerror(MAA_ERROR, "Invalid program: %s", auth_pwfile);
    return NULL;
  }

  /* set up our communication pipe */
  if (pipe (fd) == -1) {
    /* couldn't create pipe for some reason */
    ap_log_rerror (MAA_ERROR, "Error creating pipe: %s", strerror(errno));
    return NULL;
  }

   
  if(MAA_DEBUGLVL>2)
	  ap_log_rerror(MAA_DEBUG, "About to username %s to authanyprogram", user);

  /* start our child */
  pid = fork();
  switch (pid) {
  case -1:
    /* fork failed -- clean up and return */
    close (fd[0]);
    close (fd[1]);
    ap_log_rerror(MAA_ERROR, "Error fork()ing: %s", strerror(errno));
    return NULL;
    break;
  case 0:
    /* child process -- execv the authany program */
    if(MAA_DEBUGLVL>2)
	    ap_log_rerror(MAA_DEBUG, "Sending username %s to authanyprogram", user);
    child_exec(r,user,password,auth_pwfile,auth_clparams,fd);
    break;
  default:
    /* parent: read what the child has to say, if anything */
    waitpid(pid,NULL,0);
    close (fd[1]);
    do {
        i = read (fd[0], result + length, sizeof(result) - length - 1);
        if (i != -1) length += i;
    } while ((i != -1) && (i != 0) && (length < (sizeof(result) - 1)));

    /* eat up anything that might be in the pipe */
    while (0 < read(fd[0],tmpbuf,4));
    close (fd[0]);
    break;
  }

  /* chomp(result) */    
  while ((length > 0) &&
	 ((result[length - 1] == '\r') || (result[length - 1] == '\n'))) {
    result[--length] = '\0';
  }

  if ((strnlen(result,21) == 0) ||
      (strncmp(result, "Authentication Error", 19) == 0)) {
      /* not authorized */
      ap_log_rerror(MAA_ERROR, "mod_auth_any, Auth Error %s",result);
      return NULL;
  }

  /* 20031106 make sure it returned "Success\n" (followed by anything
     else, so "Success\nfoo bar baz" is ok, too. Closes bug reported by
     ali.apple.com's webmaster: if external program fails, we don't get a
     Success so we don't open ourselves. */

  if ((strncmp(result, "Success\n", 8) != 0)) {/* Success not returned */
      ap_log_rerror(MAA_ERROR, "mod_auth_any: 'Success' not returned by external authenticator");
      return NULL;
  }

  /* Chop off the "Success\n" */
  buf=result+8;

  /* save in cache */
  if(auth_cache != 0) {
      if(MAA_DEBUGLVL>0)
          ap_log_rerror(MAA_DEBUG,"updating cache with %s",user);
      TMPM_LOCK();
      update_cache(r,user,MAA_PASS,buf);
      TMPM_UNLOCK();
      if(MAA_DEBUGLVL>0)
          ap_log_rerror(MAA_DEBUG,"done updating cache with %s",user);
  }

  /* we got a winner :), return user's real name */
  /* result might be a memory leak -- caller should free it? */
  /* result is not a memory leak -- statically allocated */
  /* but a dup is returned, which should be freed */
  return strdup(buf);
} /* get_pw */

static MAA_TABLE *groups_for_user(MAA_POOL *p, char *user, char *grpfile)
{
#ifdef WITH_APACHE2
    ap_configfile_t *f;
    apr_status_t status;
#else
    configfile_t *f;
#endif
    MAA_TABLE *grps=MAA_TABLE_MAKE (p,15);
    MAA_POOL *sp;
    char l[MAX_STRING_LEN];
    const char *group_name, *ll, *w;

#ifdef WITH_APACHE2
  if ((status = ap_pcfg_openfile(&f, p, grpfile)) != APR_SUCCESS) {
#else
  if (!(f = ap_pcfg_openfile(p, grpfile))) {
#endif
    /*add?	aplog_error(APLOG_MARK, APLOG_ERR, NULL,
      "Could not open group file: %s", grpfile);*/
    return NULL;
  }

#ifdef WITH_APACHE2
  apr_pool_create(&sp, p);
#else
  sp = ap_make_sub_pool(p);
#endif

  while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
    if ((l[0] == '#') || (!l[0]))
      continue;
    ll = l;
#ifdef WITH_APACHE2
    apr_pool_clear(sp);
#else
    ap_clear_pool(sp);
#endif

    group_name = ap_getword(sp, &ll, ':');

    while (ll[0]) {
      w = ap_getword_conf(sp, &ll);
      if (!strcmp(w, user)) {
#ifdef WITH_APACHE2
          apr_table_setn(grps, apr_pstrdup(p, group_name), "in");
#else
          ap_table_setn(grps, ap_pstrdup(p, group_name), "in");
#endif
	break;
      }
    }
  }
  ap_cfg_closefile(f);
#ifdef WITH_APACHE2
  apr_pool_destroy(sp);
#else
  ap_destroy_pool(sp);
#endif
  return grps;
}

/* These functions return 0 if client is OK, and proper error status
 * if not... either AUTH_REQUIRED (HTTP_UNAUTHORIZED in APACHE2)), 
 * if we made a check, and it failed, or
 * SERVER_ERROR, if things are so totally confused that we couldn't
 * figure out how to tell if the client is authorized or not.
 *
 * If they return DECLINED, and all other modules also decline, that's
 * treated by the server core as a configuration error, logged and
 * reported as such.
 */

/* Determine user ID, and check if it really is that user, for HTTP
 * basic authentication...
 */

static int authenticate_basic_user(request_rec *r)
{
  auth_config_rec *sec =
    (auth_config_rec *) ap_get_module_config(r->per_dir_config, &auth_any_module);
  const char *sent_pw;
  char *real_pw;
  int res;
#ifdef WITH_APACHE2
  request_rec *c;
  c=r;
#else
  conn_rec *c = r->connection;
#endif
    
  if ((res = ap_get_basic_auth_pw(r, &sent_pw)))
    return res;

  if (!sec->auth_pwfile)
    return DECLINED;

  /*
    the follow get_pw() call needs to be modified to handle arbitrary
    sources for the password record.

    i'm assuming that ap_validate_password() does the actual authentication
    which is a bit of a problem since we're not necessarily maintaining
    the password in the same md5 or blowfish scheme used by apache.  and
    i don't really want to pass the entire crypt password token from the
    source.  right now i think disabling that call is our best recourse.

    Since we are relying on an external program to do validation, we
    probably can't distinguish between failure types (no such user,
    bad passwd, etc etc).  I should really handle success by evaluating
    the return code of the program, BUUUUUT I can't spend time on that right
    now.....
  */
   
  real_pw=NULL;
  if ((real_pw = get_pw(r, c->user, sent_pw, sec))==NULL) {
      if (!(sec->auth_authoritative))
          return DECLINED;
      ap_log_rerror(MAA_ERROR2,
              "user %s: Authentication failure from: %s URI: %s",
              c->user, sec->auth_pwfile, r->uri);
      ap_note_basic_auth_failure(r);
      return MAA_UNAUTH;
  }
  
  /* this is necessary. real_pw is returned by an strdup on a statically assigned array */
  if(real_pw!=NULL) {
      free(real_pw);
  }
  
  /*
    nb: eventually we will set some environment variable to reflect the token
    returned
  */
  return OK;
    
  /*
    invalid_pw = ap_validate_password(sent_pw, real_pw);
    
    if (invalid_pw != NULL) {
#ifdef WITH_APACHE2
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
    "user %s: authentication failure for \"%s\": %s",
    r->user, r->uri, invalid_pw);
    ap_note_basic_auth_failure(r);
    return HTTP_UNAUTHORIZED;
#else
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
    "user %s: authentication failure for \"%s\": %s",
    c->user, r->uri, invalid_pw);
    ap_note_basic_auth_failure(r);
    return AUTH_REQUIRED;
#endif
    }

    return OK;
  */
}

/* Checking ID */

static int check_user_access(request_rec *r)
{
  auth_config_rec *sec =
    (auth_config_rec *) ap_get_module_config(r->per_dir_config, &auth_any_module);
#ifdef WITH_APACHE2
  char *user = r->user;
  const apr_array_header_t *reqs_arr = ap_requires(r);
#else
  char *user = r->connection->user;
  const array_header *reqs_arr = ap_requires(r);
#endif
  MAA_TABLE *grpstatus;
  int m = r->method_number;
  int method_restricted = 0;
  register int x;
  const char *t, *w;
  require_line *reqs;
    
  /* BUG FIX: tadc, 11-Nov-1995.  If there is no "requires" directive, 
   * then any user will do.
   */
  if (!reqs_arr)
    return (OK);
  reqs = (require_line *) reqs_arr->elts;

  if (sec->auth_grpfile)
    grpstatus = groups_for_user(r->pool, user, sec->auth_grpfile);
  else
    grpstatus = NULL;

  for (x = 0; x < reqs_arr->nelts; x++) {

#ifdef WITH_APACHE2
    if (!(reqs[x].method_mask & (AP_METHOD_BIT << m)))
#else
    if (!(reqs[x].method_mask & (1 << m)))
#endif
      continue;

    method_restricted = 1;

    t = reqs[x].requirement;
    w = ap_getword_white(r->pool, &t);
    if (!strcmp(w, "valid-user"))
      return OK;
    if (!strcmp(w, "user")) {
      while (t[0]) {
	w = ap_getword_conf(r->pool, &t);
	if (!strcmp(user, w))
	  return OK;
      }
    }
    else if (!strcmp(w, "group")) {
      if (!grpstatus)
	return DECLINED;	/* DBM group?  Something else? */

      while (t[0]) {
	w = ap_getword_conf(r->pool, &t);
#ifdef WITH_APACHE2
    if (apr_table_get(grpstatus, w))
#else
	if (ap_table_get(grpstatus, w))
#endif
	  return OK;
      }
    } else if (sec->auth_authoritative) {
      /* if we aren't authoritative, any require directive could be
       * valid even if we don't grok it.  However, if we are 
       * authoritative, we can warn the user they did something wrong.
       * That something could be a missing "AuthAuthoritative off", but
       * more likely is a typo in the require directive.
       */
        ap_log_rerror(MAA_ERROR2,
                "access to %s failed, reason: unknown require directive:"
                "\"%s\"", r->uri, reqs[x].requirement);
    }
  }

  if (!method_restricted)
    return OK;

  if (!(sec->auth_authoritative))
    return DECLINED;

  ap_log_rerror(MAA_ERROR2,
          "access to %s failed, reason: user %s not allowed access",
          r->uri, user);
	
  ap_note_basic_auth_failure(r);
  return MAA_UNAUTH;
}


#ifdef WITH_APACHE2
static void register_hooks(MAA_POOL *p)
{
    ap_hook_check_user_id(authenticate_basic_user, NULL, NULL,APR_HOOK_MIDDLE);
    ap_hook_auth_checker(check_user_access, NULL, NULL, APR_HOOK_MIDDLE);
}
#endif


#ifdef WITH_APACHE2
module AP_MODULE_DECLARE_DATA auth_any_module =
{
    STANDARD20_MODULE_STUFF,
#else
module MODULE_VAR_EXPORT auth_any_module =
  {
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
#endif
    create_auth_dir_config,	/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    NULL,			/* server config */
    NULL,			/* merge server config */
    auth_any_cmds,		/* command table */
#ifdef WITH_APACHE2
    register_hooks              /* register hooks */
#else
    NULL,			/* handlers */
    NULL,			/* filename translation */
    authenticate_basic_user,	/* check_user_id */
    check_user_access,		/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
#endif
  };


